module Generate (..) where

import Dict exposing (Dict)
import Random exposing (Generator)
import Types exposing (..)
import Tools


initSquare : Square
initSquare =
  { isBomb = False
  , adjacentBombs = -1
  , isFlagged = False
  , isExposed = False
  , isExploded = False
  , isPressed = False
  }


initModel : Model
initModel =
  { board = Nothing
  , boardWidth = 9
  , boardHeight = 9
  , numBombs = 10
  , numFlags = 0
  , lost = False
  , numExposed = 0
  , startTime = -1
  , endTime = -1
  , generateNewBoard = Nothing
  }


countAdjacentBombs : BoardPos -> Board -> Int
countAdjacentBombs ( x, y ) board =
  Tools.countList (Maybe.withDefault False << Maybe.map (.isBomb))
    <| List.map (flip Dict.get board)
    <| List.map (\( x', y' ) -> ( x + x', y + y' ))
    <| Tools.product [ -1, 0, 1 ] [ -1, 0, 1 ]


updateAdjacentBombs : Board -> Board
updateAdjacentBombs board =
  Dict.map
    (\pos s -> { s | adjacentBombs = countAdjacentBombs pos board })
    board


boardGenerator : BoardPos -> Int -> Int -> Int -> Generator (Maybe Board)
boardGenerator ( x, y ) width height numBombs =
  let
    adjacentY =
      (if y < height - 1 then
        [ y + 1 ]
       else
        []
      )
        ++ [ y ]
        ++ (if y > 0 then
              [ y - 1 ]
            else
              []
           )

    numAdjacentY =
      List.length adjacentY

    numAdjacentX =
      1
        + (if x > 0 then
            1
           else
            0
          )
        + (if x < width - 1 then
            1
           else
            0
          )

    numAdjacent =
      numAdjacentX * numAdjacentY
  in
    if not <| 0 <= x && x < width && 0 <= y && y < height then
      Tools.constGenerator Nothing
    else
      Random.map
        (Just
          << updateAdjacentBombs
          << Dict.fromList
          << (List.map2 (,)
                <| List.map
                    (\n -> ( n % width, n // width ))
                    [0..width * height - 1]
             )
          << flip
              (List.foldr
                (\y l ->
                  Tools.insertSlice
                    (y
                      * width
                      + (if x > 0 then
                          x - 1
                         else
                          x
                        )
                    )
                    (List.repeat
                      numAdjacentX
                      initSquare
                    )
                    l
                )
              )
              adjacentY
          << List.foldr
              (\n l -> Tools.insert n { initSquare | isBomb = True } l)
              (List.repeat
                (width * height - numBombs - numAdjacent)
                initSquare
              )
        )
        <| Tools.generateList
        <| List.map
            (\n -> Random.int 0 (width * height - 1 - numAdjacent - n))
            [0..numBombs - 1]
